#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "htrack.h"

#define SOCK_NAME	"/tmp/.htrack.sock"

/* the monitor width in meters (19" LCD) */
#define MONITOR_WIDTH	0.375

/* the marker distance in meters (LEDglasses v2) */
#define MARKER_DIST		0.15

/* Logitech QuickCam 9000 Pro field of view */
#define CAM_FOV			1.0798
#define CAM_HALF_FOV	0.5399

/* consider 1u = MONITOR_WIDTH */
#define METER2UNIT(m)	((m) / MONITOR_WIDTH)
#define UNIT2METER(u)	((u) * MONITOR_WIDTH)

#define FIXED_SHIFT		15
#define FIXED_SCALE		(1 << FIXED_SHIFT)

int htrack_open(void)
{
	int s;
	struct sockaddr_un addr;

	if((s = socket(PF_UNIX, SOCK_STREAM, 0)) == -1) {
		perror("failed to create socket");
		return -1;
	}

	addr.sun_family = AF_UNIX;
	strcpy(addr.sun_path, SOCK_NAME);

	if(connect(s, (struct sockaddr*)&addr, SUN_LEN(&addr)) == -1) {
		perror("failed to connect to socket: " SOCK_NAME);
		close(s);
		return -1;
	}
	return s;
}

int htrack_close(int fd)
{
	return close(fd);
}

int htrack_read(int fd, float *x1, float *y1, float *x2, float *y2)
{
	float msg[4];
	ssize_t bytes = 0;

	do {
		if((bytes = read(fd, (char*)msg + bytes, sizeof msg - bytes)) <= 0) {
			return -1;
		}
	} while(bytes < sizeof msg);

	*x1 = msg[0];
	*y1 = msg[1];
	*x2 = msg[2];
	*y2 = msg[3];
	return 0;
}


struct point {
	float x, y;
};

#define ZSCALE	5.0
#define SQ(x)	((x) * (x))
void htrack_process(float x1, float y1, float x2, float y2, float *cam_x, float *cam_y, float *cam_z)
{
	float proj_mark_dist, view_dist, z;
	struct point p[2], mean_pt;

	p[0].x = 0.5 - x1;
	p[0].y = 0.5 - y1;
	p[1].x = 0.5 - x2;
	p[1].y = 0.5 - y2;

	mean_pt.x = (p[0].x + p[1].x) / 2.0;
	mean_pt.y = (p[0].y + p[1].y) / 2.0;

	view_dist = 0.5 / tan(CAM_HALF_FOV);
	proj_mark_dist = sqrt(SQ(p[0].x - p[1].x) + SQ(p[0].y - p[1].y));
	z = METER2UNIT(MARKER_DIST) * view_dist / proj_mark_dist;

	*cam_x = (mean_pt.x * z / view_dist) * 4.0;
	*cam_y = (mean_pt.y * z / view_dist) * 4.0;
	*cam_z = (view_dist + z * ZSCALE);
}
