#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <GL/freeglut.h>
#include <vmath/vmath.h>
#include "htrack.h"

void redraw(void);
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
unsigned int get_msec(void);
void post_redisplay(void);

int scr_xsz, scr_ysz;
int testing = 0;

float cam_x = 0;
float cam_y = 0.5;
float cam_z = 6;

float cam_tx = 0;
float cam_ty = 0.5;
float cam_tz = 0;

/* hack, see freeglut_internal.h and freeglut_init.c */
extern Display *fgDisplay;
int x11_sock;

int need_redisplay = 1;

#define NUM_TEAPOTS		40
vec3_t tpos[NUM_TEAPOTS];

#define FOV		60


int main(int argc, char **argv)
{
	int i, ht_sock;
	float lpos[] = {-10, 8, 20, 1};

	glutInitWindowSize(1024, 768);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("VR head tracking client test");
	glutFullScreen();

	scr_xsz = glutGet(GLUT_WINDOW_WIDTH);
	scr_ysz = glutGet(GLUT_WINDOW_HEIGHT);

	glutDisplayFunc(redraw);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);

	x11_sock = ConnectionNumber(fgDisplay);

	if((ht_sock = htrack_open()) == -1) {
		return 1;
	}

	for(i=0; i<NUM_TEAPOTS; i++) {
		tpos[i].x = frand(20.0) - 10.0;
		tpos[i].y = frand(10.0) - 5.0;
		tpos[i].z = frand(30.0) - 15.0;
	}

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, lpos);

	glMatrixMode(GL_PROJECTION);
	gluPerspective(FOV, (float)scr_xsz / (float)scr_ysz, 1.0, 1000.0);
	glMatrixMode(GL_MODELVIEW);

	glClearDepth(1.0);

	glNewList(1, GL_COMPILE);
	/*
	glFrontFace(GL_CW);
	glutSolidTeapot(1.0);
	glFrontFace(GL_CCW);
	*/
	glutSolidSphere(0.75, 10, 5);
	glEndList();

	for(;;) {
		int maxfd;
		fd_set rdset;

		FD_ZERO(&rdset);
		FD_SET(x11_sock, &rdset);
		FD_SET(ht_sock, &rdset);

		maxfd = x11_sock > ht_sock ? x11_sock : ht_sock;

		if(select(maxfd + 1, &rdset, 0, 0, 0) == -1) {
			if(errno != EINTR) {
				perror("select failed");
				break;
			}
			continue;
		}

		if(FD_ISSET(ht_sock, &rdset)) {
			float x1, y1, x2, y2;

			if(htrack_read(ht_sock, &x1, &y1, &x2, &y2) == -1) {
				fprintf(stderr, "connection lost");
				break;
			}
		
			htrack_process(x1, y1, x2, y2, &cam_x, &cam_y, &cam_z);
			post_redisplay();
		}

		if(FD_ISSET(x11_sock, &rdset) || need_redisplay) {
			glutMainLoopEvent();
		}
	}

	return 0;
}

void draw_grid(float spacing, float sz)
{
	float hsz = sz * 0.5;
	int i, lines = (int)(hsz / spacing) + 1;

	glLineWidth(1.0);
	glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	glColor3f(0.4, 0.4, 0.4);
	for(i=1; i<lines; i++) {
		glVertex3f(i * spacing, 0, -hsz);
		glVertex3f(i * spacing, 0, hsz);
		glVertex3f(-i * spacing, 0, -hsz);
		glVertex3f(-i * spacing, 0, hsz);
		glVertex3f(-hsz, 0, i * spacing);
		glVertex3f(hsz, 0, i * spacing);
		glVertex3f(-hsz, 0, -i * spacing);
		glVertex3f(hsz, 0, -i * spacing);
	}
	glEnd();

	glLineWidth(2.0);
	glBegin(GL_LINES);
	glColor3f(0.5, 0.5, 0.5);
	glVertex3f(-hsz, 0, 0);
	glVertex3f(hsz, 0, 0);
	glVertex3f(0, 0, -hsz);
	glVertex3f(0, 0, hsz);
	glEnd();
	glLineWidth(1.0);
	glEnable(GL_LIGHTING);
}

void draw_scene(unsigned int msec)
{
	int i;

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(cam_x, cam_y, cam_z, cam_tx, cam_ty, cam_tz, 0, 1, 0);

	for(i=0; i<NUM_TEAPOTS; i++) {
		glPushMatrix();
		glTranslatef(tpos[i].x, tpos[i].y, tpos[i].z);
		glCallList(1);
		/*glutSolidTeapot(1.0);*/
		glPopMatrix();
	}

	glPushMatrix();
	glTranslatef(0, 6, 0);
	draw_grid(4.0, 100.0);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(0, -6, 0);
	draw_grid(4.0, 100.0);
	glPopMatrix();
}

void redraw(void)
{
	unsigned int msec = get_msec();

	glClearColor(testing ? 0.8 : 0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	draw_scene(msec);

	glutSwapBuffers();
}

void reshape(int x, int y)
{
	scr_xsz = x;
	scr_ysz = y;

	glViewport(0, 0, x, y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(FOV, (float)scr_xsz / (float)scr_ysz, 1.0, 1000.0);
	glMatrixMode(GL_MODELVIEW);
}

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	default:
		break;
	}
}


unsigned int get_msec(void)
{
	static struct timeval timeval, first_timeval;
	
	gettimeofday(&timeval, 0);

	if(first_timeval.tv_sec == 0) {
		first_timeval = timeval;
		return 0;
	}
	return (timeval.tv_sec - first_timeval.tv_sec) * 1000 + (timeval.tv_usec - first_timeval.tv_usec) / 1000;
}

void post_redisplay(void)
{
	need_redisplay = 1;
	glutPostRedisplay();
}
